local TRICKSTATE_READY = 1

local BT_ACCELERATE = 1

-- Reuses some Hot Shelter assets because steam is by-and-large the same
-- This is directly copied from scripts.pk3 with slight modifications.
freeslot(
  "MT_JKDSPARK_SPAWNER",
  "S_JKDSPARK_SPAWNER",
  "MT_JKDSPARK",
  "S_JKDSPARK",
  "SPR_JKDS"
)

local function spark_spawner_thinker(mo)
  if mo.spawntimer == nil then
    mo.spawntimer = P_RandomRange(TICRATE/2, TICRATE*4)
	mo.tospawn = 0
	return
  end
  
  if mo.tospawn > 0 then
    local tospawnthisframe = P_RandomRange(1, 2)
    mo.tospawn = $ - tospawnthisframe
    for i = 1, tospawnthisframe do
	  local xyspeed = P_RandomFixed() * 4 + 2*FRACUNIT
      local variation = P_RandomRange(-2, 0)
      local variation2 = FixedAngle(P_RandomRange(-20, 20)*FRACUNIT)
      local spark = P_SpawnMobj(mo.x+2*cos(mo.angle), mo.y+2*sin(mo.angle), mo.z+1*FRACUNIT, MT_JKDSPARK)
      spark.scale = FRACUNIT
      spark.momz = variation*FRACUNIT
      spark.momx = FixedMul(xyspeed, cos(mo.angle+variation2))
      spark.momy = FixedMul(xyspeed, sin(mo.angle+variation2))
	  spark.rollangle = FixedAngle(P_RandomRange(-45, 45)*FRACUNIT)
    end
  end
  
  if mo.spawntimer > 0 then
    mo.spawntimer = $ - 1
  else
    mo.tospawn = P_RandomRange(2, 6)
	-- reset fuse
    mo.spawntimer = P_RandomRange(TICRATE/2, TICRATE*4)
  end
end

local function spark_thinker(mo)
  if mo.rotspeed == nil then
    mo.rotspeed = P_RandomRange(-5, 5) * FRACUNIT
  end
  
  mo.rollangle = $ + FixedAngle(mo.rotspeed)
  -- accelerate downward faster
  mo.momz = $ - FRACUNIT
end

mobjinfo[MT_JKDSPARK_SPAWNER] = {
  doomednum = 14301,
  spawnstate = S_JKDSPARK_SPAWNER,
  radius = 8*FRACUNIT,
  height = 8*FRACUNIT,
  flags = MF_NOCLIP|MF_NOGRAVITY,
}

states[S_JKDSPARK_SPAWNER] = {SPR_NULL, A, -1, nil, 0, 0, S_JKDSPARK_SPAWNER}

mobjinfo[MT_JKDSPARK] = {
  spawnstate = S_JKDSPARK,
  spawnhealth = 1000,
  radius = 2*FRACUNIT,
  height = 4*FRACUNIT,
  flags = MF_NOSQUISH,
}

states[S_JKDSPARK] = {SPR_JKDS, FF_ADD|FF_ANIMATE|A, 32, nil, 3, 8, S_NULL}

addHook("MobjThinker", spark_spawner_thinker, MT_JKDSPARK_SPAWNER)
addHook("MobjThinker", spark_thinker, MT_JKDSPARK)

local steamspawners = {}

-- Reuses some Hot Shelter assets because steam is by-and-large the same
-- This is directly copied from scripts.pk3 with slight modifications.
freeslot(
  "MT_JKDSTEAM_SPAWNER",
  "S_JKDSTEAM_SPAWNER"
)

local function A_KartingDrillSteam(actor, var1, var2)
  if actor.fuse <= 0 then
    actor.state = S_INVISIBLE
	return
  end
  
  actor.fuse = $ - 1

  S_StartSound(actor, sfx_lbfoom)
  local variation = P_RandomRange(-1, 1)
  local variation2 = FixedAngle(P_RandomRange(-4, 4)*FRACUNIT)
  local steam = P_SpawnMobj(actor.x+2*cos(actor.angle), actor.y+2*sin(actor.angle), actor.z+2*sin(actor.pitch), MT_HSSTEAM)
  steam.scale = FRACUNIT/4
  steam.momz = 16*sin(actor.pitch) + variation*FRACUNIT
  steam.momx = 16*cos(actor.angle+variation2)
  steam.momy = 16*sin(actor.angle+variation2)
end

mobjinfo[MT_JKDSTEAM_SPAWNER] = {
  doomednum = 14300,
  spawnstate = S_INVISIBLE,
  radius = 8*FRACUNIT,
  height = 8*FRACUNIT,
  flags = MF_NOCLIP|MF_NOGRAVITY,
}

states[S_JKDSTEAM_SPAWNER] = {SPR_NULL, A, 2, A_KartingDrillSteam, 0, 0, S_JKDSTEAM_SPAWNER}

local function clear_steam_spawners()
  -- clear steam spawners to prevent leaks
  steamspawners = {}
end

local function load_steam_spawner(mobj, spawnpoint)
  if spawnpoint.type == 14300 then
    -- add to map steam spawners and disable
	table.insert(steamspawners, mobj)
  end
end

-- Drop charge functionality
-- Actually spawns the drop charge
local function create_drop_charge(_, args)
  local player = players[args[1]]
  local dropcharge = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_DROPCHARGE)
  
  -- setup drop charge
  dropcharge.scale = 2*FRACUNIT
  -- the travel speed of the drop charge
  dropcharge.chargespeed = 40*FRACUNIT
  -- uncomment me for slower speed
  --dropcharge.chargespeed = 8*FRACUNIT
  -- the target distance of the drop charge
  --dropcharge.targetdistance = 9040*FRACUNIT
  -- unsetting targetdistance or leaving it blank causes it to run a full
  -- circuit
  dropcharge.targetdistance = nil
  
  -- enables the steam spawners for a period of time
  for i, mobj in ipairs(steamspawners) do
    mobj.fuse = 2*TICRATE
    mobj.state = S_JKDSTEAM_SPAWNER
  end
end

-- Does some changes, notably decreases all the sector brightness by 32
local function encore_mode()
  for sector in sectors.iterate do
    sector.lightlevel = $ - 32
  end
end

-- Lazy fix for bots 
local function bot_trick_doer(player, cmd)
  -- only run this hook when botstrickback is on
  if mapheaderinfo[gamemap].botstrickback then
    -- skip any spectators
    if player.spectator then return end
    if not (player.mo and player.mo.valid) then return end
  
    -- check if the bot is in a trick state
    if player.trickpanel == TRICKSTATE_READY then
      -- turn off the accelerate button
      cmd.buttons = $ & ~BT_ACCELERATE
      if player.mo.momz <= 0 then
        cmd.throwdir = -799
	    cmd.buttons = $ | BT_ACCELERATE
  	  end
    end
  end
end

addHook("MapChange", clear_steam_spawners)
addHook("MapThingSpawn", load_steam_spawner)
addHook("SpecialExecute", encore_mode, "JKDENCORE")
addHook("SpecialExecute", create_drop_charge, "JKDDISCHARGE")
addHook("BotTiccmd", bot_trick_doer)
