freeslot(
  "MT_DROPCHARGE",
  "S_DROPCHARGE",
  "MT_DROPCHARGELEAD",
  "S_DROPCHARGELEAD",
  "SPR_DRPC",
  "MT_DROPCHARGESPARK",
  "S_DROPCHARGESPARK",
  "SPR_DRPS",
  "sfx_drpcha"
)

mobjinfo[MT_DROPCHARGE] = {
  spawnstate = S_DROPCHARGE,
  spawnhealth = 1000,
  deathstate = S_NULL,
  radius = FRACUNIT,
  height = 8*FRACUNIT,
  flags = MF_NOGRAVITY,
}

states[S_DROPCHARGE] = {SPR_NULL, A, -1, nil, 0, 0, S_NULL}

mobjinfo[MT_DROPCHARGELEAD] = {
  spawnstate = S_DROPCHARGELEAD,
  spawnhealth = 1000,
  deathstate = S_NULL,
  radius = 32*FRACUNIT,
  height = FRACUNIT,
  flags = MF_NOGRAVITY|MF_PAPERCOLLISION,
}

states[S_DROPCHARGELEAD] = {SPR_DRPC, FF_ADD|FF_PAPERSPRITE|A, -1, nil, 0, 0, S_NULL}

mobjinfo[MT_DROPCHARGESPARK] = {
  spawnstate = S_DROPCHARGESPARK,
  spawnhealth = 1000,
  deathstate = S_NULL,
  radius = 32*FRACUNIT,
  height = FRACUNIT,
  flags = MF_NOGRAVITY|MF_SPECIAL,
}

states[S_DROPCHARGESPARK] = {SPR_DRPS, FF_ADD|FF_PAPERSPRITE|FF_ANIMATE|A, 15, nil, 14, 1, S_NULL}

-- setup sound effect because ring racers is stupid
-- and will not recognize SF_X8AWAYSOUND as a flag for some ungodly reason???
sfxinfo[sfx_drpcha] = {
  singular = false,
  priority = 0,
  flags = SF_X8AWAYSOUND,
}

local WP_SHORTCUT = 2
local TRICKSTATE_NONE = 0
local TRICKSTATE_READY = 1
local LOADED_WAYPOINT_VERSION = rawget(_G, "WAYPOINT_LIB_VERSION")

local PathNode = {}

PathNode.New = function(waypoint, camefrom)
  local node = {}
  node.waypoint = waypoint
  node.camefrom = camefrom
  node.totaldist = 0
  return node
end

PathNode.ClassMeta = {
  __call = function(class, ...)
    return PathNode.New(...)
  end
}

setmetatable(PathNode, PathNode.ClassMeta)

local function setup_map()
  if LOADED_WAYPOINT_VERSION == nil then
    print("\133Until 2.4 Lua expansions, the drop charge requires WaypointLib to work!")
    print("\133https://mb.srb2.org/addons/waypointutil.7793/")
	return
  end
end

-- Moves the drop charge along the waypoints
local function update_drop_charge(mo)
  -- update by last positions
  if mo.lastx ~= nil and mo.lasty ~= nil then
    local dist = R_PointToDist2(mo.lastx, mo.lasty, mo.x, mo.y)
	mo.distancetravelled = $ + dist
  end
  mo.lastx = mo.x
  mo.lasty = mo.y

  if mo.waypointpath == nil then
    return
  end
  
  local currentwaypoint
  local nextwaypoint
  local dist
  local finished = true
  if mo.targetdistance == nil or mo.distancetravelled < mo.targetdistance then
    while mo.waypointindex + 1 < #mo.waypointpath do
      currentwaypoint = mo.waypointpath[mo.waypointindex]
      nextwaypoint = mo.waypointpath[mo.waypointindex + 1]
    
      -- get distance between our position and target position
      local xydist = R_PointToDist2(mo.x, mo.y, nextwaypoint.mobj.x, nextwaypoint.mobj.y)
      dist = R_PointToDist2(0, mo.z, xydist, nextwaypoint.mobj.z)
    
      if dist <= mo.chargespeed then
        -- begin moving to the next checkpoint
	    mo.waypointindex = $ + 1
      else
	    finished = false
        break
      end
    end
  end
  
  if finished then
    -- TODO: maybe a dissipation sprite would be cool?
    mo.state = S_NULL
	return
  end
  
  -- first, get the "angle" of the waypoint before this
  local pathangle = R_PointToAngle2(currentwaypoint.mobj.x, currentwaypoint.mobj.y, nextwaypoint.mobj.x, nextwaypoint.mobj.y)
  local angleback = pathangle
  if mo.waypointindex > 1 then
    -- average the waypoint before it too
	local prevwaypoint = mo.waypointpath[mo.waypointindex - 1]
	local angle = R_PointToAngle2(prevwaypoint.mobj.x, prevwaypoint.mobj.y, currentwaypoint.mobj.x, currentwaypoint.mobj.y)
	
	local x = cos(angleback) + cos(angle)
	local y = sin(angleback) + sin(angle)
	angleback = R_PointToAngle2(0, 0, x, y)
  end
  
  -- now get the front angle
  local anglefront = pathangle
  if mo.waypointindex + 2 < #mo.waypointpath then
    -- average the waypoint after it too
	local afterwaypoint = mo.waypointpath[mo.waypointindex + 2]
	local angle = R_PointToAngle2(nextwaypoint.mobj.x, nextwaypoint.mobj.y, afterwaypoint.mobj.x, afterwaypoint.mobj.y)
	
	local x = cos(anglefront) + cos(angle)
	local y = sin(anglefront) + sin(angle)
	anglefront = R_PointToAngle2(0, 0, x, y)
  end
  
  -- find angle lerp
  local xydist = R_PointToDist2(mo.x, mo.y, currentwaypoint.mobj.x, currentwaypoint.mobj.y)
  local distback = R_PointToDist2(0, mo.z, xydist, currentwaypoint.mobj.z)
  local lerp = FixedDiv(distback, distback + dist)
  
  -- finalize angle
  mo.angle = ease.linear(lerp, 0, anglefront - angleback) + angleback
  
  -- we skip the first frame to allow the sparks to spawn
  if mo.firstframedone == nil then
    mo.firstframedone = 1
	return
  end
  
  -- normalize and multiply by speed
  mo.momx = FixedMul(FixedDiv(nextwaypoint.mobj.x - mo.x, dist), mo.chargespeed)
  mo.momy = FixedMul(FixedDiv(nextwaypoint.mobj.y - mo.y, dist), mo.chargespeed)
  mo.momz = FixedMul(FixedDiv(nextwaypoint.mobj.z - mo.z, dist), mo.chargespeed)
  
  -- play drop charge sound for all players that can still use it
  if mo.soundtimer % 21 == 0 then
    for player in players.iterate do
      if not (player and player.valid) then continue end
	  
	  local timer = mo.playersused[#player] or 0
	  if timer > 255 then continue end
	  
	  S_StartSoundAtVolume(mo, sfx_drpcha, 255 - timer, player)
    end
  end
  mo.soundtimer = $ + 1
  
  -- finally, increment the playersused timer
  for i = 0, #players - 1 do
    if mo.playersused[i] != nil then
	  mo.playersused[i] = $ + 1
	end
  end
end

local function spawn_drop_charge_spark(mo)
  local spark = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_DROPCHARGESPARK)
  spark.angle = mo.angle
  spark.target = mo.target -- set target
  
  return spark
end

-- The actual sprite of the drop charge, also spawns the interactive element
-- of the drop charge
local function update_drop_charge_lead(mo)
  if not (mo.target and mo.target.valid) then
    mo.state = S_NULL
	return
  end

  -- get position of daddy
  local targetx = mo.target.x
  local targety = mo.target.y
  local targetz = mo.target.z
  
  -- get right vector
  local angle = mo.target.angle - ANGLE_90
  mo.angle = angle
  targetx = targetx + FixedMul(mo.fracoffset, cos(angle))
  targety = targety + FixedMul(mo.fracoffset, sin(angle))
  
  -- snap to floor below charge
  targetz = P_FloorzAtPos(targetx, targety, targetz, FRACUNIT)
  
  -- teleport unit
  P_MoveOrigin(mo, targetx, targety, targetz)
  
  -- if we have travelled some distance, then spawn a charge
  if mo.sparkspawned == nil then
    -- spawn a spark as initialization
    mo.sparkspawned = 0
	spawn_drop_charge_spark(mo)
	
	-- spawn some other sparks behind this object, as it might not catch players
	-- who are going fast
	for i = 1, 2 do
	  local spark = spawn_drop_charge_spark(mo)
	  local backwards = InvAngle(mo.target.angle)
	  
	  spark.renderflags = $ | RF_DONTDRAW
	  P_MoveOrigin(spark, mo.x + cos(backwards) * i * 64, mo.y + sin(backwards) * i * 64, mo.z)
	end
  else
    local sparks = mo.target.distancetravelled / mo.sparkinterval
	if mo.sparkspawned < sparks then
	  mo.sparkspawned = sparks
	  spawn_drop_charge_spark(mo)
	end
  end
end

local function update_drop_charge_visuals(mo)
  -- we have to check if the mobj is valid because it might have been removed
  -- in a previous thinker
  if not (mo and mo.valid) then
    return
  end
  -- our target value should have the main drop charge object
  if not (mo.target and mo.target.valid) then
    return
  end
  
  local i = 0
  
  while i <= splitscreen do
    local pid = #displayplayers[i]
	
	if not mo.target.playersused[pid] then
	  break
	end
	
    i = $ + 1
  end
  
  if i > splitscreen then
    -- the charge is used up!
    mo.frame = $ | FF_TRANS70
  end
end

local function drop_charge_touch(mo, toucher)
  -- bunch of sanity checks to prevent random errors
  if not (mo and mo.valid) then return end
  if not (mo.target and mo.target.valid) then return end

  if not toucher.player then
    return true
  end
  local player = toucher.player
  
  -- check our table to see if the player has already used this drop charge
  if mo.target.playersused[#player] then
    return true
  end
  -- set this to a numeric value, which kicks off the tic timer
  mo.target.playersused[#player] = 1

  -- TRICK TIME!
  local vspeed = 20*FRACUNIT
  player.trickpanel = TRICKSTATE_READY
  player.pflags = $ | PF_TRICKDELAY
  player.tricktime = 0
  -- mark the player so they don't fall out of the trick when failing
  --player.dropchargetrick = true
  
  K_DoPogoSpring(toucher, vspeed, 0)
  S_StartSound(toucher, sfx_s3kb1, nil)
  
  -- Reduce speed of the player so they don't fling off
  --local minspeed = 24*mapobjectscale + (mapobjectscale - toucher.scale)
  --local speed = FixedHypot(mo.momx, mo.momy)
  --speed = $ / 2
  
  --if speed < minspeed then
  --  speed = minspeed
  --end
  --P_InstaThrust(mo, mo.angle, speed)

  -- do not destroy the drop charge on touch
  return true
end

local function generate_drop_charge_path(startwp)
  -- find just a path we can travel
  local openset = {}
  local closedset = {}
  local nodes = {}
  
  local startnode = PathNode(startwp)
  table.insert(openset, startnode)
  table.insert(nodes, startnode)
  
  while #openset > 0 do
    -- pop an open node
	local node = table.remove(openset)
	closedset[node] = true
	
	-- add all satisfying waypoints to the open nodes
	for i, wp in ipairs(node.waypoint.prevwaypoints) do
	  -- skip shortcuts
	  if K_GetWaypointIsShortcut(wp) then
	    continue
      end
	  
	  -- skip disabled waypoints
	  if not K_GetWaypointIsEnabled(wp) then
	    continue
	  end
	  
	  local dist = K_DistanceBetweenWaypoints(node.waypoint, wp)
	  
	  -- find node
	  local found = false
	  local found_node
	  for j, existing in ipairs(nodes) do
	    if existing.waypoint == wp then
		  found = true
		  found_node = existing
		  break
		end
	  end
	  
	  if not found then
	    local newnode = PathNode(wp, node)
	    newnode.totaldist = node.totaldist + dist
	    table.insert(openset, newnode)
		table.insert(nodes, newnode)
	  end
	end
  end
  
  -- sort by distances
  table.sort(nodes, function(a, b)
    return a.totaldist < b.totaldist
  end)
  
  -- build a path
  local current = table.remove(nodes)
  local path = {}
  while current != nil do
    table.insert(path, 1, current.waypoint)
	current = current.camefrom
  end
  
  return path
end

-- Sets up the drop charge
local function setup_drop_charge(mo)
  -- find closest waypoint to lock onto
  local closestwp = K_GetClosestWaypointToMobj(mo)
  -- lock mo
  P_MoveOrigin(mo, closestwp.mobj.x, closestwp.mobj.y, closestwp.mobj.z)
  
  local path = generate_drop_charge_path(closestwp)
  mo.waypointpath = path
  mo.waypointindex = 1
  
  -- summon the drop charge goons
  local summoncount = 7
  for i = 1, summoncount do
    local mobj = P_SpawnMobj(mo.x, mo.y, mo.z, MT_DROPCHARGELEAD)
	local offset = i - (summoncount / 2 + 1)
	-- the offset this instance is at from the main drop charge obj
	mobj.fracoffset = offset * 128 * mapobjectscale
	-- what distance needs to be travelled to spawn the spark charge
	mobj.sparkinterval = 32 * FRACUNIT
	mobj.scale = 2*mapobjectscale
	mobj.target = mo
  end
  
  -- setup distance travelled
  mo.distancetravelled = 0
  mo.soundtimer = 0
  mo.playersused = {}
  
  -- play sound
  S_StartSound(mo, sfx_s3k45, nil)
end

-- Anti frustration feature to prevent tumbles out of triangle dashes. 
/*
local function drop_charge_trick_save()
  for player in players.iterate do
    if not (player.mo and player.mo.valid and (player.trickpanel == TRICKSTATE_READY) and player.dropchargetrick and not player.spectator) continue end

    if FixedDiv(player.mo.momz * P_MobjFlip(player.mo), mapobjectscale) < -8 * FRACUNIT
      player.trickpanel = TRICKSTATE_NONE
      player.dropchargetrick = false
    end
  end
end*/

addHook("MobjSpawn", setup_drop_charge, MT_DROPCHARGE)
addHook("MobjThinker", update_drop_charge, MT_DROPCHARGE)
addHook("MobjThinker", update_drop_charge_lead, MT_DROPCHARGELEAD)
addHook("TouchSpecial", drop_charge_touch, MT_DROPCHARGESPARK)
--addHook("PreThinkFrame", drop_charge_trick_save)
addHook("MapLoad", setup_map)

addHook("MobjThinker", update_drop_charge_visuals, MT_DROPCHARGELEAD)
addHook("MobjThinker", update_drop_charge_visuals, MT_DROPCHARGESPARK)
